function [net, err] = AGD4MNIST(trainData,trainTargets,MaxIte,lrate,minibatch,NNSize,Momentum,weightdecay,initRange) 
%
%   [net, err] = SimpleNN(trainData,trainTargets) 
%
% Input: 
%       trainData(nSamples,inputSize)
%       trainTargets(nSamples,nClasses)
%       MaxIte : max # of epoch
%       lrate : learning rate
%       minibatch : minibatch size
%       NNSize: sizes of hidden layers
%       Momentum: momentum
%       weightdecay: weight decay
%       initRange: variance of initialization   
%
% Output:
%       net:   trained neural network
%       err:   training errors

%############ Topology Parameters
addpath('./Tools')
addpath('./matFiles')

load codebook.mat;
load searchList.mat;

netDesc = [1500 1000 500] ;   % # of hidden nodes in each layer
MaxEpoch = 550;       % max # of epoch in training

% fine tuning parameters
% mbsz: minibatch size; wdr: weight decay; lr: learning rate; 
% finalMomentum: momentum; initRange: variance for weight initialization
to = struct('mbsz',10,'wdr',0.000, 'lr',0.8,'finalMomentum',0, 'initRange', 0.05);


if(nargin<2 | nargin>9) 
    help SimpleNN
    return
    else

 if (nargin >= 3)
   MaxEpoch = MaxIte ;
 end
 if (nargin >= 4)
   to.lr = lrate ;
 end 
 if (nargin >= 5)
   to.mbsz = minibatch ; 
 end
 if (nargin >= 6)
   netDesc = NNSize ;
 end
 if (nargin >= 7)
   to.finalMomentum = Momentum ;
 end
 if (nargin >= 8)
   to.wdr = weightdecay ;
 end
 if (nargin >= 9)
   to.initRange = initRange ; 
 end

MaxEpoch
netDesc
to

%############ Init
trainData_Backup = trainData;
trainTargets_Backup = trainTargets;
[nSamples, inputSize] = size(trainData);
nClasses = size(trainTargets,2);

nLayers = length(netDesc);
d1 = inputSize;

matlabpool open local 4;%using multi-CPUs to parallelize image augmentation

for iLayer = 1:nLayers
  net{iLayer}.w = (rand(d1+1,netDesc(iLayer),'single')*2 - 1 ) .* to.initRange;
  
  to.dw{iLayer} = zeros(d1+1,netDesc(iLayer),'single');
  d1 = netDesc(iLayer);
  to.dw{iLayer} = GPUsingle(to.dw{iLayer});
  net{iLayer}.w = GPUsingle(net{iLayer}.w);
end
nLayers = nLayers + 1;
net{nLayers}.w = (rand(d1+1,nClasses,'single')*2 - 1) .* to.initRange ;
to.dw{nLayers} = zeros(d1+1,nClasses,'single');

net{nLayers}.w = GPUsingle(net{nLayers}.w);
to.dw{nLayers} = GPUsingle(to.dw{nLayers});
to.lr = GPUsingle(to.lr);
to.finalMomentum = GPUsingle(to.finalMomentum);

trainTargets = GPUsingle(trainTargets);


%############# Train
%batchSize = to.mbsz; %Size of training batches
batchSize = 4500; %the batchsize will start from 4500, then drop to 2500 and finally 10 in SGD
nBatches = floor(nSamples / batchSize);
k = to.wdr .* batchSize;
k = GPUsingle(k);
nNodes{1} = 785;
nNodes{2} = 1501;
nNodes{3} = 1001;
nNodes{4} = 501;
nNodes{5} = 10;
for i = 1:5
    nNodes2{i} = nNodes{i} - 1;
end
feat1{1} = ones(batchSize, nNodes{1}, GPUsingle);
feat1{2} = ones(batchSize, nNodes{2}, GPUsingle);
feat1{3} = ones(batchSize, nNodes{3}, GPUsingle);
feat1{4} = ones(batchSize, nNodes{4}, GPUsingle);

feat1_ori{1} = ones(batchSize, nNodes{1}, GPUsingle);
feat1_ori{2} = ones(batchSize, nNodes{2}, GPUsingle);
feat1_ori{3} = ones(batchSize, nNodes{3}, GPUsingle);
feat1_ori{4} = ones(batchSize, nNodes{4}, GPUsingle);

to.momentum = 0.0;
to.momentum = GPUsingle(to.momentum);

layerList = ones(nSamples, 1);
errThres = 7.5; %start with 7.5, and gradually decrease...
maxLayer = 8; %number of layers in the codebook
nFold = 1; % nFold = 1 means that we built the codebook based on the original training data. nFold more than 1 means that we firstly expand the training data by image augmentation and then build the codebook

testErr = zeros(7, MaxEpoch);
%time testErr trMSE trCE trFER
trainData_Ori = GPUsingle(trainData_Backup);
for ep = 1:MaxEpoch 
  tic
 %######################Search in the codebook to find the codeword for each training sample##############
  annIdxList = searchInBigCodeBook(trainData_Backup, trainData_Backup, nFold, inputSize);%if nFold != 1, using the expanded training set instead of the 2nd trainData_Backup
  
  if ( mean(layerList) > 5 )&&(mean(layerList) <= maxLayer)
      batchSize = 2000
      nBatches = floor(nSamples / batchSize);      
      clear feat1;  
      
      feat1{1} = ones(batchSize, nNodes{1}, GPUsingle);
      feat1{2} = ones(batchSize, nNodes{2}, GPUsingle);
      feat1{3} = ones(batchSize, nNodes{3}, GPUsingle);
      feat1{4} = ones(batchSize, nNodes{4}, GPUsingle);
	  
	  feat1_ori{1} = ones(batchSize, nNodes{1}, GPUsingle);
	  feat1_ori{2} = ones(batchSize, nNodes{2}, GPUsingle);
	  feat1_ori{3} = ones(batchSize, nNodes{3}, GPUsingle);
	  feat1_ori{4} = ones(batchSize, nNodes{4}, GPUsingle);
  elseif (mean(layerList) > maxLayer)
      batchSize = 10
	  to.lr = 0.4;
      nBatches = floor(nSamples / batchSize);     
      clear feat1;
	  
      feat1{1} = ones(batchSize, nNodes{1}, GPUsingle);
      feat1{2} = ones(batchSize, nNodes{2}, GPUsingle);
      feat1{3} = ones(batchSize, nNodes{3}, GPUsingle);
      feat1{4} = ones(batchSize, nNodes{4}, GPUsingle);
	  
	  feat1_ori{1} = ones(batchSize, nNodes{1}, GPUsingle);
	  feat1_ori{2} = ones(batchSize, nNodes{2}, GPUsingle);
	  feat1_ori{3} = ones(batchSize, nNodes{3}, GPUsingle);
	  feat1_ori{4} = ones(batchSize, nNodes{4}, GPUsingle);
  end
  
  if (mean(layerList) > maxLayer)%Do image augmentation if doing SGD
	  trainData = imageDeform(trainData_Backup);%elastic deformation
      trainData = imageRotate(trainData);
      trainData = imageScale(trainData);
      trainData = imageShear(trainData);  
	  trainData_Ann = trainData;
	  %trainData_Ann = trainData_Backup;
  else %if doing MGD, trainData_Ann contains the codewords (without image augmentation)
	  trainData_Ann = findAnnImgs(annIdxList, searchList, codebook, layerList, nSamples, inputSize, trainData_Backup, maxLayer);
  end
  
  %########Calculate the difference between original Img and the corresponding codewords#######
  errOriAnn = Euclidean_Distance(trainData_Backup, trainData_Ann);
  difErr = errThres - errOriAnn;
  difTag = find(difErr < 0);
  layerList(difTag, 1) = layerList(difTag, 1) + 1;
  reviseTmp = find(layerList > (maxLayer + 1));
  layerList(reviseTmp, 1) = maxLayer + 1;
  
  trainData = GPUsingle(trainData_Ann);
  randNdx = randperm(nSamples);
  data = slice(trainData, {randNdx}, ':');
  data_ori = slice(trainData_Ori, {randNdx}, ':');
  targets = slice(trainTargets, {randNdx}, ':');

  totMse = 0.0;
  totErr = 0.0;
  totLoss = 0.0;
  for batch = 1:nBatches
      x = slice(data, [1+(batch-1)*batchSize, 1, batch*batchSize], ':');
	  x_ori = slice(data_ori, [1+(batch-1)*batchSize, 1, batch*batchSize], ':');%x_ori will be used to calculate the network error on training set
      t = slice(targets, [1+(batch-1)*batchSize, 1, batch*batchSize], ':');

      [net, err, to, feat] = DoBackProp_GPU(x, x_ori, t, net, to, nNodes2, feat1, feat1_ori);

      totMse = totMse + err(1);
	  totErr = totErr + err(2);
      totLoss = totLoss + err(3);

  end
  ttr = toc;
  fprintf('ep %d, trCE = %f, trFER = %.2f%%, trMSE = %f, lrate =%f, in %f sec \n', ep, (totLoss*1.0/(batchSize*nBatches)), (totErr*100.0)/(batchSize*nBatches), (totMse*100)/(batchSize*nBatches), single(to.lr), ttr);
 
  save(sprintf('/tmp/network-%d',ep),'net') ;
  
  if ((totMse*100)/(batchSize*nBatches) <= 0.17)&&(ep>380)
	to.lr = single(to.lr);
	to.lr = to.lr * (10000/(10000+ep));
	to.lr = GPUsingle(to.lr);
	
  end
  for j = 1:nLayers
	net2{j}.w = single(net{j}.w);
  end
  errThres = errThres * 0.999;

  clear data data_ori targets x t trainData;

end;
matlabpool close;
save('network.mat','net');

for j = 1:nLayers
    net{j}.w = single(net{j}.w);
end

fprintf(' *********************************************************\n');

%############## training error
[err] = TestNet_GPU(trainData_Backup, trainTargets_Backup, net);

fprintf(' *********************************************************\n');


end
